% Inference example doing sparse regression for image deblurring using glm-ie.
%
% We do nonblind deconvolution to recover the image u that has been blurred by
% an optical device with known point spread function f.
% The measurements are y = conv2(u,f) + e, e~N(0,s2).
%
% (c) by Hannes Nickisch, MPI for Biological Cybernetics, 2010 November 26
clear all, close all

load deblur
ut = u; clear u, f = f(4:end,:); f = f/sum(f(:));    % image u and blur filter f
su = size(ut); nu = prod(su); sf = size(f);                              % sizes

% blur operator and noisy measurement
X  = matConv2(f, su, 'circ'); sy = get_sout(X);
s2 = 1e-5;                                                            % variance
y  = X*ut; y = y + sqrt(s2)*randn(size(y));                        % observation

% prior
B = [matWav(su); matFD2(su)]; % wavelet and total variation = finite diff matrix
tau = 15;                                                          % prior scale

pot = @potLaplace;

%% Inference
opts.innerOutput = 1;
opts.outerOutput = 1;
opts.outerNiter = 4;          % Number of outer loop iterations
opts.outerMVM   = 170;        % Lanczos steps
opts.innerMVM   = 100;        % Number of CG steps in inner loop
opts.innerVBpls = 'plsLBFGS'; % PLS algorithm for inner loop
[uinf,ga,b,z,nlZ,Q,T] = dli(X,y,s2,B,pot,tau,opts); uinf = reshape(uinf,su);
% compute standard deviation: sdinf = sqrt(s2*diag( Q*inv(T)*Q' ));
% V = inv(A), A = X'*X/s2 + B'*diag(1./ga)*B \approx Q*T*Q'
Q_T = Q/chol(T);
sdinf = sqrt( sum(Q_T.*Q_T,2) ); sdinf = reshape(sdinf,su);

%% Estimation
opt.nMVM = 250; opt.output = 1;
u0 = zeros(nu,1);
uest = reshape( feval(opts.innerVBpls,u0,X,y,B,opt,s2,'penVB',pot,tau), su);

% accuracy
err_estimation = norm(uest(:)-ut(:))
err_inference  = norm(uinf(:)-ut(:))

% di-f1.png
sz = [800,300]; figure('Position',[50,50,sz])
subplot('Position',[0/3,0,.32,.87])
  imagesc(abs(uest-ut),[0,.2]),  axis off, title('|true-mode|','FontSize',16)
subplot('Position',[1/3,0,.32,.87])
  imagesc(abs(uinf-ut),[0,.2]),  axis off, title('|true-mean|','FontSize',16)
  subplot('Position',[2/3,0,.32,.87])
  imagesc(sdinf,[0,3e-3]),       axis off, title('standard dev.','FontSize',16)
colormap(gray)